import numpy as np
import pandas as pd
import math

days_in_year = 250
default_risk_free_rate = 0.05


class RatioCalculator:
    def __init__(
        self,
        nav_dataframe,
        benchmark_nav_dataframe=None,
        risk_free_rate=None,
        annualiser=None,
    ):
        self_nav_df = self._load(nav_dataframe)
        self.risk_free_rate = risk_free_rate
        self.annualiser = annualiser if annualiser else days_in_year
        if benchmark_nav_dataframe is None:
            benchmark_nav_dataframe = self.create_benchmark_nav_from_risk_free_rate(
                nav_dataframe, risk_free_rate
            )
        benchmark_nav_df = self._load(benchmark_nav_dataframe)
        self.combo_nav_df = self._merge(self_nav_df, benchmark_nav_df)
        self.beta_cache = None

    @staticmethod
    def _load(data,):
        if data is None:
            return pd.DataFrame()
        ret = "returns"
        data[ret] = data["nav"].pct_change()

        return data

    def create_benchmark_nav_from_risk_free_rate(self, nav_dataframe, risk_free_rate):
        if not risk_free_rate:
            return None
        initial_value = 100
        risk_free_rate = risk_free_rate / self.annualiser
        benchmark_nav_df = nav_dataframe.filter(["nav"])
        initial_index = benchmark_nav_df.index.min()

        def apply_drag(row, initial_value, initial_index, rate):
            count = (row.name - initial_index).days
            return initial_value * math.pow(1 + rate, count)

        benchmark_nav_df["nav"] = benchmark_nav_df.apply(
            apply_drag, args=(initial_value, initial_index, risk_free_rate), axis=1
        )
        return benchmark_nav_df.filter(["nav"])

    @staticmethod
    def _merge(scheme_df, benchmark_df):
        if not benchmark_df.empty:
            combo_df = scheme_df.join(benchmark_df, how="left", rsuffix="_benchmark")
            return combo_df
        return scheme_df

    def get_treynor(self, period=None):
        beta_df = self.get_beta(period)
        df = self.combo_nav_df
        df = df.join(beta_df, how="inner", rsuffix="_beta")
        df["treynor"] = (df["returns"] - df["returns_benchmark"]) / df["beta"]
        return df.filter(["treynor"])

    def get_alpha(self, window):
        beta_df = self.get_beta(window)

        df = self.combo_nav_df

        df = df.join(beta_df, how="inner")
        df["cumulative_returns"] = df["returns"].cumsum()
        df["cumulative_returns_benchmark"] = df["returns_benchmark"].cumsum()
        df["alpha"] = df["cumulative_returns"] - (
            df["cumulative_returns_benchmark"]
            + df["beta"]
            * (df["cumulative_returns"] - df["cumulative_returns_benchmark"])
        )
        return df.filter(["alpha"])

    def get_beta(self, window):
        """
        https://stackoverflow.com/questions/39501277/efficient-python-pandas-stock-beta-calculation-on-many-dataframes

        beta = covariance(returns, benchmark returns) / variance(benchmark returns)
        also excel cov(a, b) = b.cov(a) in pandas
        """
        if self.beta_cache is not None:
            return self.beta_cache

        df = self.combo_nav_df.copy()
        df["var"] = df["returns_benchmark"].rolling(window=window).var()
        df["cov"] = df["returns_benchmark"].rolling(window=window).cov(df["returns"])
        df["beta"] = df["cov"] / df["var"]
        df = df.filter(["beta"])
        self.beta_cache = df
        return df

    def get_upside_capture(self, window):
        """
        requires benchmark data
        """

        df = self.combo_nav_df
        df["scheme_return_when_benchmark_up"] = df["returns"].where(
            df["returns_benchmark"] > 0, 0
        )
        df["benchmark_return_when_benchmark_up"] = df["returns_benchmark"].where(
            df["returns_benchmark"] > 0, 0
        )

        df["upside_cagr_fund"] = (1 + df["scheme_return_when_benchmark_up"]).rolling(
            window=window
        ).apply(np.prod, raw=True) - 1
        df["upside_cagr_index"] = (
            1 + df["benchmark_return_when_benchmark_up"]
        ).rolling(window=window).apply(np.prod, raw=True) - 1

        df["upside_capture_ratio"] = (1 + df["upside_cagr_fund"]) / (
            1 + df["upside_cagr_index"]
        )
        return df

    def get_downside_capture(self, window):
        """
        requires benchmark data
        """
        df = self.combo_nav_df
        df["scheme_return_when_benchmark_down"] = df["returns"].where(
            df["returns_benchmark"] < 0, 0
        )
        df["benchmark_return_when_benchmark_down"] = df["returns_benchmark"].where(
            df["returns_benchmark"] < 0, 0
        )

        df["downside_cagr_fund"] = (
            1 + df["scheme_return_when_benchmark_down"]
        ).rolling(window=window).apply(np.prod, raw=True) - 1
        df["downside_cagr_index"] = (
            1 + df["benchmark_return_when_benchmark_down"]
        ).rolling(window=window).apply(np.prod, raw=True) - 1

        df["downside_capture_ratio"] = (1 - df["downside_cagr_fund"]) / (
            1 - df["downside_cagr_index"]
        )
        return df

    def get_drawdown(self, window):
        """
        here window means that a max would be taken for that rolling window
        :param period:
        :return:
        """
        df = self.combo_nav_df

        df["nav_peak"] = df["nav"].rolling(window=window).max()
        df["drawdown"] = (df["nav"] / df["nav_peak"]) - 1

        # how to get days
        # df = df.tail(window)
        # max_drawdown_idx = df["drawdown"].idxmin()
        # max_drawdown_idx_peak = df["nav_peak"][max_drawdown_idx]
        # for i, r in df.iterrows():
        #     if i > max_drawdown_idx and r["nav"] >= max_drawdown_idx_peak:
        #         days = (i-max_drawdown_idx).days
        #         print(days)
        #         break

        return df

    def get_volatility(self, window):
        df = self.combo_nav_df
        df["volatility"] = df["returns"].rolling(window=window).std() * math.sqrt(
            self.annualiser
        )
        return df.filter(["volatility"])

    def get_sharpe(self, window):
        """
        https://stackoverflow.com/questions/49091044/python-rolling-sharpe-ratio-with-pandas-or-numpy
        """
        #   (rp - rf) / sigma p
        #   mean(per date returns row) - mean(per date benchmark return)
        #   -----------------------------------------------------------
        #                   std(excess returns)
        #

        df = self.combo_nav_df

        df["returns_mean"] = (
            (1 + df["returns"].rolling(window=window).mean()) ** self.annualiser
        ) - 1
        df["benchmark_returns_mean"] = (
            (1 + df["returns_benchmark"].rolling(window=window).mean())
            ** self.annualiser
        ) - 1
        df["excess_returns_std"] = (df["returns"] - df["returns_benchmark"]).rolling(
            window=window
        ).std() * math.sqrt(self.annualiser)
        df["sharpe"] = (df["returns_mean"] - df["benchmark_returns_mean"]) / df[
            "excess_returns_std"
        ]

        return df.filter(["sharpe"])

    def get_sortino(self, window):
        """
        Similar to sharpe, only negative returns are considered
        """
        df = self.combo_nav_df

        df["returns_mean"] = (
            (1 + df["returns"].rolling(window=window).mean()) ** self.annualiser
        ) - 1
        df["benchmark_returns_mean"] = (
            (1 + df["returns_benchmark"].rolling(window=window).mean())
            ** self.annualiser
        ) - 1
        df["excess_returns"] = df["returns"] - df["returns_benchmark"]
        df["excess_returns_downside"] = df["excess_returns"].where(
            df["excess_returns"] < 0, 0
        )
        df["downside_excess_returns_std"] = df["excess_returns_downside"].rolling(
            window=window
        ).std() * math.sqrt(self.annualiser)

        df["sortino"] = (df["returns_mean"] - df["benchmark_returns_mean"]) / df[
            "downside_excess_returns_std"
        ]

        return df.filter(["sortino"])



# # import scipy.optimize
# from datetime import date


# def xnpv(rate, cashflows):
#     """
#     Equivalent of Excel's XNPV function.
#     >>> from datetime import date
#     >>> dates = [date(2010, 12, 29), date(2012, 1, 25), date(2012, 3, 8)]
#     >>> values = [-10000, 20, 10100]
#     >>> xnpv(0.1, values, dates)
#     -966.4345...
#     """
#     if rate <= -1.0:
#         return float('inf')
#     d0 = cashflows[0][0]    # or min(dates)
#     return sum([c[1] / (1.0 + rate)**((c[0] - d0).days / 365.0) for c in cashflows])
#
#
# def xirr(cashflows):
#     """
#     Equivalent of Excel's XIRR function.
#     >>> from datetime import date
#     >>> dates = [date(2010, 12, 29), date(2012, 1, 25), date(2012, 3, 8)]
#     >>> values = [-10000, 20, 10100]
#     >>> xirr(values, dates)
#     0.0100612...
#     """
#     try:
#         return scipy.optimize.newton(lambda r: xnpv(r, cashflows), 0.0)
#     except RuntimeError:    # Failed to converge?
#         return scipy.optimize.brentq(lambda r: xnpv(r, cashflows), -1.0, 1e10)


# noinspection PyBroadException


class XIRR:
    # Set maximum epsilon for end of iteration
    eps_max_rate = 1e-6
    eps_max_value = 1e-4
    guess_vals_neg = [-0.1, -0.05, -0.025, -0.15, -0.2, -0.25, -0.3, -0.5, -0.75, -0.9]

    # Set maximum epsilon for end of iteration
    iter_max = 100

    def __init__(self, transactions, guess=None):
        self.transactions = transactions
        self.sort_transactions()
        self.xirr = None
        self.guess = guess
        self.error = False

    def sort_transactions(self):
        try:
            self.transactions.sort(key=lambda x: [x[0], -x[1]])
        except TypeError:
            self.error = True

    def check_if_correct_transactions(self):
        # Check that values contains at least one positive value and one negative value
        positive = False
        negative = False
        # todo: scope of improvement
        for t in self.transactions:
            if t[1] > 0:
                positive = True
            if t[1] < 0:
                negative = True
            if positive and negative:
                break
        if positive and negative:
            correct = True
        else:
            correct = False
        return correct

    def set_guess_rate(self):
        if self.guess is None:
            if sum(k[1] for k in self.transactions) > 0:
                self.guess = -0.1
            else:
                self.guess = 0.1

    def get_xirr(self):
        transactions_correct = self.check_if_correct_transactions()
        if transactions_correct and not self.error:

            guesses_tried_pos = list(map(lambda x: -x, self.guess_vals_neg))
            self.set_guess_rate()
            if self.guess == guesses_tried_pos[0]:
                guesses_try = guesses_tried_pos
            else:
                guesses_try = self.guess_vals_neg
            guesses_try.reverse()
            tries = 0
            while tries < 1 or (self.xirr is None and len(guesses_try)):
                self.guess = guesses_try.pop()
                self.xirr = self.calculate_xirr()
                tries += 1
        return self.xirr

    def calculate_xirr(self):
        try:
            result_rate = self.implement_newtons_method()
            if type(result_rate) not in [float, int]:
                return None
        except (ZeroDivisionError, OverflowError, TypeError):
            result_rate = None
        except Exception:
            result_rate = None
        return result_rate

    def implement_newtons_method(self):
        result_rate = self.guess
        # Implement Newton's method
        iteration = 0
        cont_loop = True
        while cont_loop and (iteration < self.iter_max):
            # Result  value  gives you residual value from the assumed rate of return
            result_value = self.irr_result(result_rate)
            new_rate = result_rate - (result_value / self.irr_result_deriv(result_rate))

            eps_rate = abs(new_rate - result_rate)
            result_rate = new_rate

            iteration += 1
            cont_loop = (eps_rate > self.eps_max_rate) and (
                abs(result_value) > self.eps_max_value
            )
        if cont_loop:
            result_rate = None
        else:
            result_rate = result_rate * 100
        return result_rate

    def irr_result(self, rate):
        r = rate + 1
        result = 0
        first = self.transactions[0]

        for t in self.transactions:
            frac = (t[0] - first[0]).days / 365
            result += t[1] / pow(r, frac)
        return result

    def irr_result_deriv(self, rate):
        # Calculates the first derivation
        r = rate + 1
        result = 0
        first = self.transactions[0]
        for t in self.transactions:
            frac = (t[0] - first[0]).days / 365
            result -= frac * t[1] / pow(r, frac + 1)
        return result

    def set_guess_for_extreme_cases(self):
        pos_amt = sum(k[1] for k in self.transactions if k[1] > 0)
        neg_amt = abs(sum(k[1] for k in self.transactions if k[1] < 0))
        per = (self.transactions[-1][0] - self.transactions[0][0]).days / 365
        self.guess = pow(neg_amt / pos_amt, 1 / per) - 1



# flake8: noqa
from .api import (
    get_sortino,
    get_xirr,
    get_sharpe,
    get_volatility,
    get_drawdown,
    get_downside_capture,
    get_upside_capture,
    get_alpha,
    get_treynor,
    get_beta,
)

__version__ = "__version__ = '0.0.5'"
